@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2015 Con Kolivas
+   Copyright (C) 2006-2016 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998-2003 Andrew Tridgell
 
@@ -311,8 +311,8 @@ int open_tmpoutfile(rzip_control *control)
 
 	fd_out = mkstemp(control->outfile);
 	if (fd_out == -1) {
-		print_verbose("WARNING: Failed to create out tmpfile: %s , will fail if cannot perform entirely in ram\n",
-			      control->outfile, DECOMPRESS ? "de" : "");
+		print_progress("WARNING: Failed to create out tmpfile: %s, will fail if cannot perform %scompression entirely in ram\n",
+			       control->outfile, DECOMPRESS ? "de" : "");
 	} else
 		register_outfile(control, control->outfile, TEST_ONLY || STDOUT || !KEEP_BROKEN);
 	return fd_out;
@@ -424,31 +424,50 @@ bool write_fdin(rzip_control *control)
 /* Open a temporary inputfile to perform stdin decompression */
 int open_tmpinfile(rzip_control *control)
 {
-	int fd_in;
+	int fd_in = -1;
 
+	/* Use temporary directory if there is one */
 	if (control->tmpdir) {
 		control->infile = malloc(strlen(control->tmpdir) + 15);
 		if (unlikely(!control->infile))
 			fatal_return(("Failed to allocate infile name\n"), -1);
 		strcpy(control->infile, control->tmpdir);
 		strcat(control->infile, "lrzipin.XXXXXX");
-	} else {
-		control->infile = malloc(15);
+		fd_in = mkstemp(control->infile);
+	}
+
+	/* Try the current directory */
+	if (fd_in == -1) {
+		free(control->infile);
+		control->infile = malloc(16);
 		if (unlikely(!control->infile))
 			fatal_return(("Failed to allocate infile name\n"), -1);
 		strcpy(control->infile, "lrzipin.XXXXXX");
+		fd_in = mkstemp(control->infile);
 	}
 
-	fd_in = mkstemp(control->infile);
-	if (unlikely(fd_in == -1))
-		fatal_return(("Failed to create in tmpfile: %s\n", control->infile), -1);
-	register_infile(control, control->infile, (DECOMPRESS || TEST_ONLY) && STDIN);
-	/* Unlink temporary file immediately to minimise chance of files left
-	 * lying around in cases of failure_return((. */
-	if (unlikely(unlink(control->infile))) {
-		fatal("Failed to unlink tmpfile: %s\n", control->infile);
-		close(fd_in);
-		return -1;
+	/* Use /tmp if nothing is writeable so far */
+	if (fd_in == -1) {
+		free(control->infile);
+		control->infile = malloc(20);
+		if (unlikely(!control->infile))
+			fatal_return(("Failed to allocate infile name\n"), -1);
+		strcpy(control->infile, "/tmp/lrzipin.XXXXXX");
+		fd_in = mkstemp(control->infile);
+	}
+
+	if (fd_in == -1) {
+		print_progress("WARNING: Failed to create in tmpfile: %s, will fail if cannot perform %scompression entirely in ram\n",
+			       control->infile, DECOMPRESS ? "de" : "");
+	} else {
+		register_infile(control, control->infile, (DECOMPRESS || TEST_ONLY) && STDIN);
+		/* Unlink temporary file immediately to minimise chance of files left
+		* lying around in cases of failure_return((. */
+		if (unlikely(unlink(control->infile))) {
+			fatal("Failed to unlink tmpfile: %s\n", control->infile);
+			close(fd_in);
+			return -1;
+		}
 	}
 	return fd_in;
 }
@@ -474,6 +493,8 @@ bool read_tmpinfile(rzip_control *control, int fd_in)
 	FILE *tmpinfp;
 	int tmpchar;
 
+	if (fd_in == -1)
+		return false;
 	if (control->flags & FLAG_SHOW_PROGRESS)
 		fprintf(control->msgout, "Copying from stdin.\n");
 	tmpinfp = fdopen(fd_in, "w+");
@@ -492,13 +513,25 @@ bool read_tmpinfile(rzip_control *control, int fd_in)
  * a pseudo-temporary file */
 static bool open_tmpoutbuf(rzip_control *control)
 {
+	i64 maxlen = control->maxram;
+	void *buf;
+
+	while (42) {
+		round_to_page(&maxlen);
+		buf = malloc(maxlen);
+		if (buf) {
+			print_maxverbose("Malloced %"PRId64" for tmp_outbuf\n", maxlen);
+			break;
+		}
+		maxlen = maxlen / 3 * 2;
+		if (maxlen < 100000000)
+			fatal_return(("Unable to even malloc 100MB for tmp_outbuf\n"), false);
+	}
 	control->flags |= FLAG_TMP_OUTBUF;
-	control->out_maxlen = control->maxram;
 	/* Allocate slightly more so we can cope when the buffer overflows and
 	 * fall back to a real temporary file */
-	control->tmp_outbuf = malloc(control->maxram + control->page_size);
-	if (unlikely(!control->tmp_outbuf))
-		fatal_return(("Failed to malloc tmp_outbuf in open_tmpoutbuf\n"), false);
+	control->out_maxlen = maxlen - control->page_size;
+	control->tmp_outbuf = buf;
 	if (!DECOMPRESS && !TEST_ONLY)
 		control->out_ofs = control->out_len = MAGIC_LEN;\
 	return true;
@@ -703,8 +736,6 @@ bool decompress_file(rzip_control *control)
 
 	if (STDIN) {
 		fd_in = open_tmpinfile(control);
-		if (unlikely(fd_in == -1))
-			return false;
 		read_tmpinmagic(control);
 		if (ENCRYPT)
 			failure_return(("Cannot decompress encrypted file from STDIN\n"), false);
@@ -757,9 +788,10 @@ bool decompress_file(rzip_control *control)
 	if (unlikely(!open_tmpoutbuf(control)))
 		return false;
 
-	if (!STDIN)
+	if (!STDIN) {
 		if (unlikely(!read_magic(control, fd_in, &expected_size)))
 			return false;
+	}
 
 	if (!STDOUT && !TEST_ONLY) {
 		/* Check if there's enough free space on the device chosen to fit the
@@ -802,7 +834,7 @@ bool decompress_file(rzip_control *control)
 	/* if we get here, no fatal_return(( errors during decompression */
 	print_progress("\r");
 	if (!(STDOUT | TEST_ONLY))
-		print_output("Output filename is: %s: ", control->outfile);
+		print_progress("Output filename is: %s: ", control->outfile);
 	if (!expected_size)
 		expected_size = control->st_size;
 	if (!ENCRYPT)
@@ -820,7 +852,7 @@ bool decompress_file(rzip_control *control)
 
 	close(fd_in);
 
-	if (!KEEP_FILES) {
+	if (!KEEP_FILES && !STDIN) {
 		if (unlikely(unlink(control->infile)))
 			fatal_return(("Failed to unlink %s\n", infilecopy), false);
 	}
@@ -929,7 +961,8 @@ bool get_fileinfo(rzip_control *control)
 	infile_size = st.st_size;
 
 	/* Get decompressed size */
-	if (unlikely(!read_magic(control, fd_in, &expected_size))) goto error;
+	if (unlikely(!read_magic(control, fd_in, &expected_size)))
+		goto error;
 
 	if (ENCRYPT) {
 		print_output("Encrypted lrzip archive. No further information available\n");
@@ -1189,10 +1222,14 @@ bool compress_file(rzip_control *control)
 			fatal_goto(("Failed to create %s\n", control->outfile), error);
 		}
 		control->fd_out = fd_out;
-		if (!STDIN)
-			if (unlikely(!preserve_perms(control, fd_in, fd_out))) goto error;
-	} else
-		if (unlikely(!open_tmpoutbuf(control))) goto error;
+		if (!STDIN) {
+			if (unlikely(!preserve_perms(control, fd_in, fd_out)))
+				goto error;
+		}
+	} else {
+		if (unlikely(!open_tmpoutbuf(control)))
+			goto error;
+	}
 
 	/* Write zeroes to header at beginning of file */
 	if (unlikely(!STDOUT && write(fd_out, header, sizeof(header)) != sizeof(header)))
@@ -1201,12 +1238,19 @@ bool compress_file(rzip_control *control)
 	rzip_fd(control, fd_in, fd_out);
 
 	/* Wwrite magic at end b/c lzma does not tell us properties until it is done */
-	if (!STDOUT)
-		if (unlikely(!write_magic(control))) goto error;
+	if (!STDOUT) {
+		if (unlikely(!write_magic(control)))
+			goto error;
+	}
 
 	if (ENCRYPT)
 		release_hashes(control);
 
+	if (unlikely(!STDIN && !STDOUT && !preserve_times(control, fd_in))) {
+		fatal("Failed to preserve times on output file\n");
+		goto error;
+	}
+
 	if (unlikely(close(fd_in))) {
 		fatal("Failed to close fd_in\n");
 		fd_in = -1;
@@ -1217,7 +1261,7 @@ bool compress_file(rzip_control *control)
 	if (TMP_OUTBUF)
 		close_tmpoutbuf(control);
 
-	if (!KEEP_FILES) {
+	if (!KEEP_FILES && !STDIN) {
 		if (unlikely(unlink(control->infile)))
 			fatal_return(("Failed to unlink %s\n", control->infile), false);
 	}
@@ -1225,14 +1269,16 @@ bool compress_file(rzip_control *control)
 	free(control->outfile);
 	return true;
 error:
-	if (STDIN && (fd_in > 0)) close(fd_in);
-	if ((!STDOUT) && (fd_out > 0)) close(fd_out);
+	if (STDIN && (fd_in > 0))
+		close(fd_in);
+	if ((!STDOUT) && (fd_out > 0))
+		close(fd_out);
 	return false;
 }
 
 bool initialise_control(rzip_control *control)
 {
-	struct timeval tv;
+	time_t now_t, tdiff;
 	char *eptr; /* for environment */
 
 	memset(control, 0, sizeof(rzip_control));
@@ -1253,9 +1299,17 @@ bool initialise_control(rzip_control *control)
 	/* The first 5 bytes of the salt is the time in seconds.
 	 * The next 2 bytes encode how many times to hash the password.
 	 * The last 9 bytes are random data, making 16 bytes of salt */
-	if (unlikely(gettimeofday(&tv, NULL)))
-		fatal_return(("Failed to gettimeofday in main\n"), false);
-	control->secs = tv.tv_sec;
+	if (unlikely((now_t = time(NULL)) == ((time_t)-1)))
+		fatal_return(("Failed to call time in main\n"), false);
+	if (unlikely(now_t < T_ZERO)) {
+		print_output("Warning your time reads before the year 2011, check your system clock\n");
+		now_t = T_ZERO;
+	}
+	/* Workaround for CPUs no longer keeping up with Moore's law!
+	 * This way we keep the magic header format unchanged. */
+	tdiff = (now_t - T_ZERO) / 4;
+	now_t = T_ZERO + tdiff;
+	control->secs = now_t;
 	control->encloops = nloops(control->secs, control->salt, control->salt + 1);
 	if (unlikely(!get_rand(control, control->salt + 2, 6)))
 		return false;
@@ -1276,13 +1330,13 @@ bool initialise_control(rzip_control *control)
 	}
 	size_t len = strlen(eptr);
 
-	control->tmpdir = malloc(len+2);
+	control->tmpdir = malloc(len + 2);
 	if (control->tmpdir == NULL)
 		fatal_return(("Failed to allocate for tmpdir\n"), false);
 	strcpy(control->tmpdir, eptr);
-	if (control->tmpdir[len] != '/')
+	if (control->tmpdir[len - 1] != '/') {
 		control->tmpdir[len] = '/'; /* need a trailing slash */
-	control->tmpdir[len+1] = '\0';
-
+		control->tmpdir[len + 1] = '\0';
+	}
 	return true;
 }
