@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2013 Con Kolivas
+   Copyright (C) 2006-2016 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998-2003 Andrew Tridgell
 
@@ -30,6 +30,7 @@
 #include <stdio.h>
 #include <stdbool.h>
 #include <stdarg.h>
+#include <semaphore.h>
 
 #ifdef HAVE_PTHREAD_H
 # include <pthread.h>
@@ -66,6 +67,10 @@ void *alloca (size_t);
 # include <sys/endian.h>
 #endif
 #ifndef __BYTE_ORDER
+# ifndef __BIG_ENDIAN
+#  define __BIG_ENDIAN	4321
+#  define __LITTLE_ENDIAN	1234
+# endif
 # ifdef WORDS_BIGENDIAN
 #  define __BYTE_ORDER __BIG_ENDIAN
 # else
@@ -143,12 +148,24 @@ extern int errno;
 # define ffsll __builtin_ffsll
 #endif
 
-typedef long long int i64;
+typedef int64_t i64;
 typedef uint32_t u32;
 
 typedef struct rzip_control rzip_control;
 typedef struct md5_ctx md5_ctx;
 
+/* ck specific unnamed semaphore implementations to cope with osx not
+ * implementing them. */
+#ifdef __APPLE__
+struct cksem {
+	int pipefd[2];
+};
+
+typedef struct cksem cksem_t;
+#else
+typedef sem_t cksem_t;
+#endif
+
 #if !defined(__linux)
  #define mremap fake_mremap
 #endif
@@ -325,6 +342,36 @@ struct checksum {
 	i64 len;
 };
 
+typedef i64 tag;
+
+struct rzip_state {
+	void *ss;
+	struct level *level;
+	tag hash_index[256];
+	struct hash_entry *hash_table;
+	char hash_bits;
+	i64 hash_count;
+	i64 hash_limit;
+	tag minimum_tag_mask;
+	i64 tag_clean_ptr;
+	i64 last_match;
+	i64 chunk_size;
+	i64 mmap_size;
+	char chunk_bytes;
+	uint32_t cksum;
+	int fd_in, fd_out;
+	char stdin_eof;
+	struct {
+		i64 inserts;
+		i64 literals;
+		i64 literal_bytes;
+		i64 matches;
+		i64 match_bytes;
+		i64 tag_hits;
+		i64 tag_misses;
+	} stats;
+};
+
 struct rzip_control {
 	char *infile;
 	FILE *inFILE; // if a FILE is being read from
@@ -379,7 +426,7 @@ struct rzip_control {
 	unsigned char magic_written;
 	bool lzma_prop_set;
 
-	pthread_mutex_t cksumlock;
+	cksem_t cksumsem;
 	md5_ctx ctx;
 	uchar md5_resblock[MD5_DIGEST_SIZE];
 	i64 md5_read; // How far into the file the md5 has done so far
@@ -398,12 +445,15 @@ struct rzip_control {
 	int log_level;
 	void (*info_cb)(void *data, int pct, int chunk_pct);
 	void *info_data;
-	void (*log_cb)(void *data, unsigned int level, unsigned int line, const char *file, const char *func, const char *format, va_list);
+	void (*log_cb)(void *data, unsigned int level, unsigned int line, const char *file, const char *func, const char *format, va_list args);
 	void *log_data;
 
+	char chunk_bytes;
 	struct sliding_buffer sb;
-	uchar *(*get_sb)(rzip_control *control, i64 p);
-	void (*do_mcpy)(rzip_control *control, unsigned char *buf, i64 offset, i64 len);
+	void (*do_mcpy)(rzip_control *, unsigned char *, i64, i64);
+	void (*next_tag)(rzip_control *, struct rzip_state *, i64, tag *);
+	tag (*full_tag)(rzip_control *, struct rzip_state *, i64);
+	i64 (*match_len)(rzip_control *, struct rzip_state *, i64, i64, i64, i64 *);
 };
 
 struct stream {
@@ -445,8 +495,8 @@ static inline void print_stuff(const rzip_control *control, int level, unsigned
 	} else if (control->msgout) {
 		va_start(ap, format);
 		vfprintf(control->msgout, format, ap);
-		fflush(control->msgout);
 		va_end(ap);
+		fflush(control->msgout);
 	}
 }
 
@@ -461,6 +511,7 @@ static inline void print_err(const rzip_control *control, unsigned int line, con
 		va_start(ap, format);
 		vfprintf(control->msgerr, format, ap);
 		va_end(ap);
+		fflush(control->msgerr);
 	}
 }
 
