@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2013 Con Kolivas
+   Copyright (C) 2006-2016 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998-2003 Andrew Tridgell
 
@@ -49,6 +49,10 @@
 # include <arpa/inet.h>
 #endif
 
+#include <dirent.h>
+#include <getopt.h>
+#include <libgen.h>
+
 #include "rzip.h"
 #include "lrzip_core.h"
 #include "util.h"
@@ -57,47 +61,70 @@
 /* needed for CRC routines */
 #include "lzma/C/7zCrc.h"
 
+#define MAX_PATH_LEN 4096
+
 static rzip_control base_control, local_control, *control;
 
-static void usage(void)
+static void usage(bool compat)
 {
-	print_output("lrzip version %s\n", PACKAGE_VERSION);
-	print_output("Copyright (C) Con Kolivas 2006-2013\n");
+	print_output("lrz%s version %s\n", compat ? "" : "ip", PACKAGE_VERSION);
+	print_output("Copyright (C) Con Kolivas 2006-2016\n");
 	print_output("Based on rzip ");
 	print_output("Copyright (C) Andrew Tridgell 1998-2003\n\n");
-	print_output("Usage: lrzip [options] <file...>\n");
+	print_output("Usage: lrz%s [options] <file...>\n", compat ? "" : "ip");
 	print_output("General options:\n");
-	print_output("     -c            check integrity of file written on decompression\n");
-	print_output("     -d            decompress\n");
-	print_output("     -e            password protected sha512/aes128 encryption on compression\n");
-	print_output("     -h|-?         show help\n");
-	print_output("     -H            display md5 hash integrity information\n");
-	print_output("     -i            show compressed file information\n");
-	print_output("     -q            don't show compression progress\n");
-	print_output("     -t            test compressed file integrity\n");
-	print_output("     -v[v]         Increase verbosity\n");
-	print_output("     -V            show version\n");
+	if (compat) {
+		print_output("	-c, --stdout		output to STDOUT\n");
+		print_output("	-C, --check		check integrity of file written on decompression\n");
+	} else
+		print_output("	-c, -C, --check		check integrity of file written on decompression\n");
+	print_output("	-d, --decompress	decompress\n");
+	print_output("	-e, --encrypt		password protected sha512/aes128 encryption on compression\n");
+	print_output("	-h, -?, --help		show help\n");
+	print_output("	-H, --hash		display md5 hash integrity information\n");
+	print_output("	-i, --info		show compressed file information\n");
+	if (compat) {
+		print_output("	-L, --license		display software version and license\n");
+		print_output("	-P, --progress		show compression progress\n");
+	} else
+		print_output("	-q, --quiet		don't show compression progress\n");
+	print_output("	-r, --recursive		operate recursively on directories\n");
+	print_output("	-t, --test		test compressed file integrity\n");
+	print_output("	-v[v%s], --verbose	Increase verbosity\n", compat ? "v" : "");
+	print_output("	-V, --version		show version\n");
 	print_output("Options affecting output:\n");
-	print_output("     -D            delete existing files\n");
-	print_output("     -f            force overwrite of any existing files\n");
-	print_output("     -k            keep broken or damaged output files\n");
-	print_output("     -o filename   specify the output file name and/or path\n");
-	print_output("     -O directory  specify the output directory when -o is not used\n");
-	print_output("     -S suffix     specify compressed suffix (default '.lrz')\n");
+	if (!compat)
+		print_output("	-D, --delete		delete existing files\n");
+	print_output("	-f, --force		force overwrite of any existing files\n");
+	if (compat)
+		print_output("	-k, --keep		don't delete source files on de/compression\n");
+	print_output("	-K, --keep-broken	keep broken or damaged output files\n");
+	print_output("	-o, --outfile filename	specify the output file name and/or path\n");
+	print_output("	-O, --outdir directory	specify the output directory when -o is not used\n");
+	print_output("	-S, --suffix suffix	specify compressed suffix (default '.lrz')\n");
 	print_output("Options affecting compression:\n");
-	print_output("     -b            bzip2 compression\n");
-	print_output("     -g            gzip compression using zlib\n");
-	print_output("     -l            lzo compression (ultra fast)\n");
-	print_output("     -n            no backend compression - prepare for other compressor\n");
-	print_output("     -z            zpaq compression (best, extreme compression, extremely slow)\n");
+	print_output("	--lzma			lzma compression (default)\n");
+	print_output("	-b, --bzip2		bzip2 compression\n");
+	print_output("	-g, --gzip		gzip compression using zlib\n");
+	print_output("	-l, --lzo		lzo compression (ultra fast)\n");
+	print_output("	-n, --no-compress	no backend compression - prepare for other compressor\n");
+	print_output("	-z, --zpaq		zpaq compression (best, extreme compression, extremely slow)\n");
 	print_output("Low level options:\n");
-	print_output("     -L level      set lzma/bzip2/gzip compression level (1-9, default 7)\n");
-	print_output("     -N value      Set nice value to value (default 19)\n");
-	print_output("     -p value      Set processor count to override number of threads\n");
-	print_output("     -T            Disable LZO compressibility testing\n");
-	print_output("     -U            Use unlimited window size beyond ramsize (potentially much slower)\n");
-	print_output("     -w size       maximum compression window in hundreds of MB\n");
-	print_output("                   default chosen by heuristic dependent on ram and chosen compression\n");
+	if (compat) {
+		print_output("	-1 .. -9		set lzma/bzip2/gzip compression level (1-9, default 7)\n");
+		print_output("	--fast			alias for -1\n");
+		print_output("	--best			alias for -9\n");
+	}
+	if (!compat)
+		print_output("	-L, --level level	set lzma/bzip2/gzip compression level (1-9, default 7)\n");
+	print_output("	-N, --nice-level value	Set nice value to value (default %d)\n", compat ? 0 : 19);
+	print_output("	-p, --threads value	Set processor count to override number of threads\n");
+	print_output("	-m, --maxram size	Set maximim available ram in hundreds of MB\n");
+	print_output("				overrides detected ammount of available ram\n");
+	print_output("	-T, --threshold		Disable LZO compressibility testing\n");
+	print_output("	-U, --unlimited		Use unlimited window size beyond ramsize (potentially much slower)\n");
+	print_output("	-w, --window size	maximum compression window in hundreds of MB\n");
+	print_output("				default chosen by heuristic dependent on ram and chosen compression\n");
 	print_output("\nLRZIP=NOCONFIG environment variable setting can be used to bypass lrzip.conf.\n");
 	print_output("TMP environment variable will be used for storage of temporary files when needed.\n");
 	print_output("TMPDIR may also be stored in lrzip.conf file.\n");
@@ -105,6 +132,17 @@ static void usage(void)
 
 }
 
+static void license(void)
+{
+	print_output("lrz version %s\n", PACKAGE_VERSION);
+	print_output("Copyright (C) Con Kolivas 2006-2016\n");
+	print_output("Based on rzip ");
+	print_output("Copyright (C) Andrew Tridgell 1998-2003\n\n");
+	print_output("This is free software.  You may redistribute copies of it under the terms of\n");
+	print_output("the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.\n");
+	print_output("There is NO WARRANTY, to the extent permitted by law.\n");
+}
+
 static void sighandler(int sig __UNUSED__)
 {
 	struct termios termios_p;
@@ -184,26 +222,118 @@ static void show_summary(void)
 	}
 }
 
+static struct option long_options[] = {
+	{"bzip2",	no_argument,	0,	'b'}, /* 0 */
+	{"check",	no_argument,	0,	'c'},
+	{"check",	no_argument,	0,	'C'},
+	{"decompress",	no_argument,	0,	'd'},
+	{"delete",	no_argument,	0,	'D'},
+	{"encrypt",	no_argument,	0,	'e'}, /* 5 */
+	{"force",	no_argument,	0,	'f'},
+	{"gzip",	no_argument,	0,	'g'},
+	{"help",	no_argument,	0,	'h'},
+	{"hash",	no_argument,	0,	'H'},
+	{"info",	no_argument,	0,	'i'}, /* 10 */
+	{"keep-broken",	no_argument,	0,	'k'},
+	{"keep-broken",	no_argument,	0,	'K'},
+	{"lzo",		no_argument,	0,	'l'},
+	{"lzma",       	no_argument,	0,	'/'},
+	{"level",	optional_argument,	0,	'L'}, /* 15 */
+	{"license",	no_argument,	0,	'L'},
+	{"maxram",	required_argument,	0,	'm'},
+	{"no-compress",	no_argument,	0,	'n'},
+	{"nice-level",	required_argument,	0,	'N'},
+	{"outfile",	required_argument,	0,	'o'},
+	{"outdir",	required_argument,	0,	'O'}, /* 20 */
+	{"threads",	required_argument,	0,	'p'},
+	{"progress",	no_argument,	0,	'P'},
+	{"quiet",	no_argument,	0,	'q'},
+	{"recursive",	no_argument,	0,	'r'},
+	{"suffix",	required_argument,	0,	'S'},
+	{"test",	no_argument,	0,	't'},  /* 25 */
+	{"threshold",	required_argument,	0,	'T'},
+	{"unlimited",	no_argument,	0,	'U'},
+	{"verbose",	no_argument,	0,	'v'},
+	{"version",	no_argument,	0,	'V'},
+	{"window",	required_argument,	0,	'w'},  /* 30 */
+	{"zpaq",	no_argument,	0,	'z'},
+	{"fast",	no_argument,	0,	'1'},
+	{"best",	no_argument,	0,	'9'},
+	{0,	0,	0,	0},
+};
+
+static void set_stdout(struct rzip_control *control)
+{
+	control->flags |= FLAG_STDOUT;
+	control->outFILE = stdout;
+	control->msgout = stderr;
+	register_outputfile(control, control->msgout);
+}
+
+/* Recursively enter all directories, adding all regular files to the dirlist array */
+static void recurse_dirlist(char *indir, char **dirlist, int *entries)
+{
+	char fname[MAX_PATH_LEN];
+	struct stat istat;
+	struct dirent *dp;
+	DIR *dirp;
+
+	dirp = opendir(indir);
+	if (unlikely(!dirp))
+		failure("Unable to open directory %s\n", indir);
+	while ((dp = readdir(dirp)) != NULL) {
+		if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, ".."))
+			continue;
+		sprintf(fname, "%s/%s", indir, dp->d_name);
+		if (unlikely(stat(fname, &istat)))
+			failure("Unable to stat file %s\n", fname);
+		if (S_ISDIR(istat.st_mode)) {
+			recurse_dirlist(fname, dirlist, entries);
+			continue;
+		}
+		if (!S_ISREG(istat.st_mode)) {
+			print_err("Not regular file %s\n", fname);
+			continue;
+		}
+		print_maxverbose("Added file %s\n", fname);
+		*dirlist = realloc(*dirlist, MAX_PATH_LEN * (*entries + 1));
+		strcpy(*dirlist + MAX_PATH_LEN * (*entries)++, fname);
+	}
+	closedir(dirp);
+}
+
+static const char *loptions = "bcCdDefghHiKlL:nN:o:O:p:PqrS:tTUm:vVw:z?";
+static const char *coptions = "bcCdefghHikKlLnN:o:O:p:PrS:tTUm:vVw:z?123456789";
+
 int main(int argc, char *argv[])
 {
+	bool lrzcat = false, compat = false, recurse = false;
 	struct timeval start_time, end_time;
 	struct sigaction handler;
 	double seconds,total_time; // for timers
-	bool lrzcat = false;
 	int c, i;
 	int hours,minutes;
 	extern int optind;
-	char *eptr; /* for environment */
+	char *eptr, *av; /* for environment */
 
         control = &base_control;
 
-	initialize_control(control);
+	initialise_control(control);
 
-	if (strstr(argv[0], "lrunzip"))
+	av = basename(argv[0]);
+	if (!strcmp(av, "lrunzip"))
 		control->flags |= FLAG_DECOMPRESS;
-	else if (strstr(argv[0], "lrzcat")) {
+	else if (!strcmp(av, "lrzcat")) {
 		control->flags |= FLAG_DECOMPRESS | FLAG_STDOUT;
 		lrzcat = true;
+	} else if (!strcmp(av, "lrz")) {
+		/* Called in gzip compatible command line mode */
+		control->flags &= ~FLAG_SHOW_PROGRESS;
+		control->nice_val = 0;
+		control->flags &= ~FLAG_KEEP_FILES;
+		compat = true;
+		long_options[1].name = "stdout";
+		long_options[11].name = "keep";
 	}
 
 	/* generate crc table */
@@ -219,7 +349,7 @@ int main(int argc, char *argv[])
 	else if (!strstr(eptr,"NOCONFIG"))
 		read_config(control);
 
-	while ((c = getopt(argc, argv, "bcdDefghHiklL:nN:o:O:p:qS:tTUvVw:z?")) != -1) {
+	while ((c = getopt_long(argc, argv, compat ? coptions : loptions, long_options, &i)) != -1) {
 		switch (c) {
 		case 'b':
 			if (control->flags & FLAG_NOT_LZMA)
@@ -227,6 +357,12 @@ int main(int argc, char *argv[])
 			control->flags |= FLAG_BZIP2_COMPRESS;
 			break;
 		case 'c':
+			if (compat) {
+				control->flags |= FLAG_KEEP_FILES;
+				set_stdout(control);
+				break;
+			}
+		case 'C':
 			control->flags |= FLAG_CHECK;
 			control->flags |= FLAG_HASH;
 			break;
@@ -249,7 +385,7 @@ int main(int argc, char *argv[])
 			break;
 		case 'h':
 		case '?':
-			usage();
+			usage(compat);
 			return -1;
 		case 'H':
 			control->flags |= FLAG_HASH;
@@ -258,6 +394,11 @@ int main(int argc, char *argv[])
 			control->flags |= FLAG_INFO;
 			break;
 		case 'k':
+			if (compat) {
+				control->flags |= FLAG_KEEP_FILES;
+				break;
+			}
+		case 'K':
 			control->flags |= FLAG_KEEP_BROKEN;
 			break;
 		case 'l':
@@ -266,10 +407,17 @@ int main(int argc, char *argv[])
 			control->flags |= FLAG_LZO_COMPRESS;
 			break;
 		case 'L':
+			if (compat) {
+				license();
+				exit(0);
+			}
 			control->compression_level = atoi(optarg);
 			if (control->compression_level < 1 || control->compression_level > 9)
 				failure("Invalid compression level (must be 1-9)\n");
 			break;
+		case 'm':
+			control->ramsize = atol(optarg) * 1024 * 1024 * 100;
+			break;
 		case 'n':
 			if (control->flags & FLAG_NOT_LZMA)
 				failure("Can only use one of -l, -b, -g, -z or -n\n");
@@ -305,9 +453,15 @@ int main(int argc, char *argv[])
 			if (control->threads < 1)
 				failure("Must have at least one thread\n");
 			break;
+		case 'P':
+			control->flags |= FLAG_SHOW_PROGRESS;
+			break;
 		case 'q':
 			control->flags &= ~FLAG_SHOW_PROGRESS;
 			break;
+		case 'r':
+			recurse = true;
+			break;
 		case 'S':
 			if (control->outname)
 				failure("Specified output filename already, can't specify an extension.\n");
@@ -318,6 +472,8 @@ int main(int argc, char *argv[])
 		case 't':
 			if (control->outname)
 				failure("Cannot specify an output file name when just testing.\n");
+			if (compat)
+				control->flags |= FLAG_KEEP_FILES;
 			if (!KEEP_FILES)
 				failure("Doubt that you want to delete a file when just testing.\n");
 			control->flags |= FLAG_TEST_ONLY;
@@ -330,7 +486,9 @@ int main(int argc, char *argv[])
 			break;
 		case 'v':
 			/* set verbosity flag */
-			if (!(control->flags & FLAG_VERBOSITY) && !(control->flags & FLAG_VERBOSITY_MAX))
+			if (!(control->flags & FLAG_SHOW_PROGRESS))
+				control->flags |= FLAG_SHOW_PROGRESS;
+			else if (!(control->flags & FLAG_VERBOSITY) && !(control->flags & FLAG_VERBOSITY_MAX))
 				control->flags |= FLAG_VERBOSITY;
 			else if ((control->flags & FLAG_VERBOSITY)) {
 				control->flags &= ~FLAG_VERBOSITY;
@@ -349,14 +507,29 @@ int main(int argc, char *argv[])
 				failure("Can only use one of -l, -b, -g, -z or -n\n");
 			control->flags |= FLAG_ZPAQ_COMPRESS;
 			break;
+		case '1':
+		case '2':
+		case '3':
+		case '4':
+		case '5':
+		case '6':
+		case '7':
+		case '8':
+		case '9':
+			control->compression_level = c - '0';
+			break;
 		}
 	}
 
 	argc -= optind;
 	argv += optind;
 
-	if (control->outname && argc > 1)
-		failure("Cannot specify output filename with more than 1 file\n");
+	if (control->outname) {
+		if (argc > 1)
+			failure("Cannot specify output filename with more than 1 file\n");
+		if (recurse)
+			failure("Cannot specify output filename with recursive\n");
+	}
 
 	if (VERBOSE && !SHOW_PROGRESS) {
 		print_err("Cannot have -v and -q options. -v wins.\n");
@@ -390,35 +563,55 @@ int main(int argc, char *argv[])
 
 	/* One extra iteration for the case of no parameters means we will default to stdin/out */
 	for (i = 0; i <= argc; i++) {
+		char *dirlist = NULL, *infile = NULL;
+		int direntries = 0, curentry = 0;
+
 		if (i < argc)
-			control->infile = argv[i];
+			infile = argv[i];
 		else if (!(i == 0 && STDIN))
 			break;
-		if (control->infile) {
-			if ((strcmp(control->infile, "-") == 0))
+		if (infile) {
+			if ((strcmp(infile, "-") == 0))
 				control->flags |= FLAG_STDIN;
 			else {
-				struct stat infile_stat;
+				bool isdir = false;
+				struct stat istat;
 
-				stat(control->infile, &infile_stat);
-				if (unlikely(S_ISDIR(infile_stat.st_mode)))
-					failure("lrzip only works directly on FILES.\n"
-					"Use lrztar or pipe through tar for compressing directories.\n");
+				if (unlikely(stat(infile, &istat)))
+					failure("Failed to stat %s\n", infile);
+				isdir = S_ISDIR(istat.st_mode);
+				if (!recurse && (isdir || !S_ISREG(istat.st_mode))) {
+					failure("lrzip only works directly on regular FILES.\n"
+					"Use -r recursive, lrztar or pipe through tar for compressing directories.\n");
+				}
+				if (recurse && !isdir)
+					failure("%s not a directory, -r recursive needs a directory\n", infile);
 			}
 		}
 
+		if (recurse) {
+			if (unlikely(STDIN || STDOUT))
+				failure("Cannot use -r recursive with STDIO\n");
+			recurse_dirlist(infile, &dirlist, &direntries);
+		}
+
 		if (INFO && STDIN)
 			failure("Will not get file info from STDIN\n");
-
-		if ((control->outname && (strcmp(control->outname, "-") == 0)) ||
-			/* If no output filename is specified, and we're using
-			 * stdin, use stdout */
-			(!control->outname && STDIN) || lrzcat ) {
-				control->flags |= FLAG_STDOUT;
-				control->outFILE = stdout;
-				control->msgout = stderr;
-				register_outputfile(control, control->msgout);
+recursion:
+		if (recurse) {
+			if (curentry >= direntries) {
+				infile = NULL;
+				continue;
+			}
+			infile = dirlist + MAX_PATH_LEN * curentry++;
 		}
+		control->infile = infile;
+
+		/* If no output filename is specified, and we're using
+		 * stdin, use stdout */
+		if ((control->outname && (strcmp(control->outname, "-") == 0)) ||
+		    (!control->outname && STDIN) || lrzcat)
+				set_stdout(control);
 
 		if (lrzcat) {
 			control->msgout = stderr;
@@ -434,6 +627,8 @@ int main(int argc, char *argv[])
 		if (STDIN)
 			control->inFILE = stdin;
 		/* Implement signal handler only once flags are set */
+		sigemptyset(&handler.sa_mask);
+		handler.sa_flags = 0;
 		handler.sa_handler = &sighandler;
 		sigaction(SIGTERM, &handler, 0);
 		sigaction(SIGINT, &handler, 0);
@@ -441,12 +636,12 @@ int main(int argc, char *argv[])
 		if (!FORCE_REPLACE) {
 			if (STDIN && isatty(fileno((FILE *)stdin))) {
 				print_err("Will not read stdin from a terminal. Use -f to override.\n");
-				usage();
+				usage(compat);
 				exit (1);
 			}
-			if (!TEST_ONLY && STDOUT && isatty(fileno((FILE *)stdout))) {
+			if (!TEST_ONLY && STDOUT && isatty(fileno((FILE *)stdout)) && !compat) {
 				print_err("Will not write stdout to a terminal. Use -f to override.\n");
-				usage();
+				usage(compat);
 				exit (1);
 			}
 		}
@@ -486,6 +681,8 @@ int main(int argc, char *argv[])
 		seconds = total_time - hours * 3600 - minutes * 60;
 		if (!INFO)
 			print_progress("Total time: %02d:%02d:%05.2f\n", hours, minutes, seconds);
+		if (recurse)
+			goto recursion;
 	}
 
 	return 0;
